from setuptools_cpp import CMakeExtension, ExtensionBuilder
from typing import Any, Dict


def build(setup_kwargs: Dict[str, Any]) -> None:
    ext_modules = [
        CMakeExtension("llamacpp.llamacpp", sourcedir="."),
    ]

    setup_kwargs.update(
        {
            "ext_modules": ext_modules,
            "cmdclass": dict(build_ext=ExtensionBuilder),
            "zip_safe": False,
        }
    )



# Convert a LLaMA model checkpoint to a ggml compatible file
#
# Load the model using Torch
# Iterate over all variables and write them to a binary file.
#
# For each variable, write the following:
#   - Number of dimensions (int)
#   - Name length (int)
#   - Dimensions (int[n_dims])
#   - Name (char[name_length])
#   - Data (float[n_dims])
#
# By default, the bigger matrices are converted to 16-bit floats.
# This can be disabled by adding the "use-f32" CLI argument.
#
# At the start of the ggml file we write the model parameters
# and vocabulary.
#

import sys
import json
import struct
import numpy as np
import torch

from sentencepiece import SentencePieceProcessor

def main():
    if len(sys.argv) < 3:
        print("Usage: convert-ckpt-to-ggml.py dir-model ftype\n")
        print("  ftype == 0 -> float32")
        print("  ftype == 1 -> float16")
        sys.exit(1)

    # output in the same directory as the model
    dir_model = sys.argv[1]

    fname_hparams   = sys.argv[1] + "/params.json"
    fname_tokenizer = sys.argv[1] + "/../tokenizer.model"

    def get_n_parts(dim):
        if dim == 4096:
            return 1
        elif dim == 5120:
            return 2
        elif dim == 6656:
            return 4
        elif dim == 8192:
            return 8
        else:
            print("Invalid dim: " + str(dim))
            sys.exit(1)

    # possible data types
    #   ftype == 0 -> float32
    #   ftype == 1 -> float16
    #
    # map from ftype to string
    ftype_str = ["f32", "f16"]

    ftype = 1
    if len(sys.argv) > 2:
        ftype = int(sys.argv[2])
        if ftype < 0 or ftype > 1:
            print("Invalid ftype: " + str(ftype))
            sys.exit(1)
        fname_out = sys.argv[1] + "/ggml-model-" + ftype_str[ftype] + ".bin"

    with open(fname_hparams, "r") as f:
        hparams = json.load(f)

    tokenizer = SentencePieceProcessor(fname_tokenizer)

    hparams.update({"vocab_size": tokenizer.vocab_size()})

    n_parts = get_n_parts(hparams["dim"])

    print(hparams)
    print('n_parts = ', n_parts)

    for p in range(n_parts):
        print('Processing part ', p)

        #fname_model = sys.argv[1] + "/consolidated.00.pth"
        fname_model = sys.argv[1] + "/consolidated.0" + str(p) + ".pth"
        fname_out = sys.argv[1] + "/ggml-model-" + ftype_str[ftype] + ".bin"
        if (p > 0):
            fname_out = sys.argv[1] + "/ggml-model-" + ftype_str[ftype] + ".bin" + "." + str(p)

        # weights_only requires torch 1.13.1, remove this param or update if you get an "invalid keyword argument" error
        model = torch.load(fname_model, map_location="cpu", weights_only=True)

        fout = open(fname_out, "wb")

        fout.write(struct.pack("i", 0x67676d6c)) # magic: ggml in hex
        fout.write(struct.pack("i", hparams["vocab_size"]))
        fout.write(struct.pack("i", hparams["dim"]))
        fout.write(struct.pack("i", hparams["multiple_of"]))
        fout.write(struct.pack("i", hparams["n_heads"]))
        fout.write(struct.pack("i", hparams["n_layers"]))
        fout.write(struct.pack("i", hparams["dim"] // hparams["n_heads"])) # rot (obsolete)
        fout.write(struct.pack("i", ftype))

        # Is this correct??
        for i in range(32000):
            # TODO: this is probably wrong - not sure how this tokenizer works
            text = tokenizer.decode([29889, i]).encode('utf-8')
            # remove the first byte (it's always '.')
            text = text[1:]
            fout.write(struct.pack("i", len(text)))
            fout.write(text)

        for k, v in model.items():
            name = k
            shape = v.shape

            # skip layers.X.attention.inner_attention.rope.freqs
            if name[-5:] == "freqs":
                continue

            print("Processing variable: " + name + " with shape: ", shape, " and type: ", v.dtype)

            #data = tf.train.load_variable(dir_model, name).squeeze()
            data = v.numpy().squeeze()
            n_dims = len(data.shape);

            # for efficiency - transpose some matrices
            # "model/h.*/attn/c_attn/w"
            # "model/h.*/attn/c_proj/w"
            # "model/h.*/mlp/c_fc/w"
            # "model/h.*/mlp/c_proj/w"
            #if name[-14:] == "/attn/c_attn/w" or \
            #   name[-14:] == "/attn/c_proj/w" or \
            #   name[-11:] == "/mlp/c_fc/w" or \
            #   name[-13:] == "/mlp/c_proj/w":
            #    print("  Transposing")
            #    data = data.transpose()

            dshape = data.shape

            # default type is fp16
            ftype_cur = 1
            if ftype == 0 or n_dims == 1:
                print("  Converting to float32")
                data = data.astype(np.float32)
                ftype_cur = 0

            # header
            sname = name.encode('utf-8')
            fout.write(struct.pack("iii", n_dims, len(sname), ftype_cur))
            for i in range(n_dims):
                fout.write(struct.pack("i", dshape[n_dims - 1 - i]))
            fout.write(sname);

            # data
            data.tofile(fout)

        # I hope this deallocates the memory ..
        model = None

        fout.close()

        print("Done. Output file: " + fname_out + ", (part ", p, ")")
        print("")

if __name__ == '__main__':
    main()



"""Python version of main.cpp"""
import sys
import argparse
import llamacpp


def parse_args_into_params(argv) -> llamacpp.gpt_params:
    """Parses arguments using argparse based on usage information above"""
    parser = argparse.ArgumentParser(description="llama.cpp CLI")
    parser.add_argument("-i", "--interactive", action="store_true", help="run in interactive mode")
    parser.add_argument(
        "--interactive-start",
        action="store_true",
        help="run in interactive mode and poll user input at startup",
        default=False,
    )
    parser.add_argument(
        "-r",
        "--reverse-prompt",
        type=str,
        help="in interactive mode, poll user input upon seeing PROMPT",
        default="",
    )
    parser.add_argument(
        "--color",
        action="store_true",
        help="colorise output to distinguish prompt and user input from generations",
    )
    parser.add_argument("-s", "--seed", type=int, default=-1, help="RNG seed (default: -1)")
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=4,
        help="number of threads to use during computation (default: 1)",
    )
    parser.add_argument(
        "-p",
        "--prompt",
        type=str,
        help="prompt to start generation with (default: random)",
        required=True,
    )
    # parser.add_argument(
    #     "-f", "--file", type=str, default="", help="prompt file to start generation."
    # )
    parser.add_argument(
        "-n", "--n_predict", type=int, default=128, help="number of tokens to predict (default: 128)"
    )
    parser.add_argument("--top_k", type=int, default=40, help="top-k sampling (default: 40)")
    parser.add_argument("--top_p", type=float, default=0.95, help="top-p sampling (default: 0.1)")
    parser.add_argument(
        "--repeat_last_n",
        type=int,
        default=64,
        help="last n tokens to consider for penalize (default: 0)",
    )
    parser.add_argument(
        "--repeat_penalty",
        type=float,
        default=1.30,
        help="penalize repeat sequence of tokens (default: 0.0)",
    )
    parser.add_argument(
        "-c",
        "--ctx_size",
        type=int,
        default=4096,
        help="size of the prompt context (default: 4096)",
    )
    parser.add_argument("--temp", type=float, default=0.8, help="temperature (default: 0.7)")
    parser.add_argument(
        "-b",
        "--batch_size",
        type=int,
        default=8,
        help="batch size for prompt processing (default: 2)",
    )
    parser.add_argument("-m", "--model", type=str, default="./models/7B/ggml-model-q4_0.bin", help="model path (default: )")
    parser.usage = parser.format_help()

    args = parser.parse_args(argv[1:])

    # Add a space in front of the first character to match OG llama tokenizer behavior
    args.prompt = " " + args.prompt

    # Initialize gpt_params object
    params = llamacpp.gpt_params(
        args.model,
        args.prompt,
        args.reverse_prompt,
        args.ctx_size,
        args.n_predict,
        args.top_k,
        args.top_p,
        args.temp,
        args.repeat_penalty,
        args.seed,
        args.threads,
        args.repeat_last_n,
        args.batch_size,
        args.color,
        args.interactive or args.interactive_start,
        args.interactive_start,
    )

    return params


def process_interactive_input(model: llamacpp.PyLLAMA):
    """Process interactive input similar to the C++ version"""

    # Read lines as long as user is entering "\" at the end of the line
    # Pass each line to the model
    while True:
        line = input()
        if line.endswith("\\"):
            line = line[:-1]
            model.update_input(line)
        else:
            model.update_input(line)
            break


def main(params):
    """Main function"""
    model = llamacpp.PyLLAMA(params)
    model.add_bos()
    model.update_input(params.prompt)
    model.print_startup_stats()
    model.prepare_context()

    # Set antiprompt if we are in interactive mode
    if params.interactive:
        model.set_antiprompt(params.antiprompt)

    if params.interactive:
        print("== Running in interactive mode. ==")
        print(" - Press Ctrl+C to interject at any time.")
        print(" - Press Return to return control to LLaMa.")
        print(" - If you want to submit another line, end your input in '\\'.")
        print()

    # prompt user immediately after the starting prompt has been loaded
    if params.interactive_start:
        is_interacting = True

    input_noecho = False
    is_finished = False

    while not model.is_finished():
        if model.has_unconsumed_input():
            model.ingest_all_pending_input(not input_noecho)
            # # reset color to default if we there is no pending user input
            # if (!input_noecho && params.use_color) {
            #     printf(ANSI_COLOR_RESET);
            # }
        else:
            text, is_finished = model.infer_text()
            print(text, end="")
            input_noecho = False

        if params.interactive:
            if model.is_antiprompt_present():
                # reverse prompt found
                is_interacting = True
            if is_interacting:
                process_interactive_input(model)
                input_noecho = True
                is_interacting = False

        if is_finished:
            break

    return 0


def run():
    # Parse params into a gpt_params object
    params = parse_args_into_params(sys.argv)
    return main(params)

if __name__ == "__main__":
    sys.exit(run())



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['llamacpp']

package_data = \
{'': ['*']}

install_requires = \
['sentencepiece>=0.1.97,<0.2.0', 'torch>=1.13.1,<2.0.0']

entry_points = \
{'console_scripts': ['llamacpp-chat = llamacpp.chat:run',
                     'llamacpp-cli = llamacpp.cli:run',
                     'llamacpp-convert = llamacpp.convert:main',
                     'llamacpp-quantize = llamacpp.quantize:main']}

setup_kwargs = {
    'name': 'llamacpp',
    'version': '0.1.5',
    'description': "Python bindings for @ggerganov's llama.cpp",
    'long_description': '## Python bindings for llama.cpp\n\n## Building the Python bindings\n\n### macOS\n\n```\nbrew install pybind11  # Installs dependency\ngit submodule init && git submodule update\npoetry install\n```\n### From PyPI\n\n```\npip install llamacpp\n```\n\n## Get the model weights\n\nYou will need to obtain the weights for LLaMA yourself. There are a few torrents floating around as well as some huggingface repositories (e.g https://huggingface.co/nyanko7/LLaMA-7B/). Once you have them, copy them into the models folder.\n\n```\nls ./models\n65B 30B 13B 7B tokenizer_checklist.chk tokenizer.model\n```\n\nConvert the weights to GGML format using `llamacpp-convert`. Then use `llamacpp-quantize` to quantize them into INT4. For example, for the 7B parameter model, run\n\n```\nllamacpp-convert ./models/7B/ 1\nllamacpp-quantize ./models/7B/\nllamacpp-cli\n```\n\n## Command line interface\n\nThe package installs the command line entry point `llamacpp-cli` that points to `llamacpp/cli.py` and should provide about the same functionality as the `main` program in the original C++ repository. There is also an experimental `llamacpp-chat` that is supposed to bring up a chat interface but this is not working correctly yet.\n\n## Demo script\n\nSee `llamacpp/cli.py` for a detailed example. The simplest demo would be something like the following:\n\n## ToDo\n\n- [x] Use poetry to build package\n- [x] Add command line entry point for quantize script\n- [x] Publish wheel to PyPI\n- [ ] Add chat interface based on tinygrad\n',
    'author': 'Thomas Antony',
    'author_email': 'mail@thomasantony.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/thomasantony/llamacpp-python',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
